@@ -1,5 +1,14 @@
+#!/usr/bin/env python
+# -*- coding: utf-8 -*-
+# @Date    : 2020-05-17
+# @Author  : Shawn Shan (shansixiong@cs.uchicago.edu)
+# @Link    : https://www.shawnshan.com/
+
+
+import errno
 import glob
 import gzip
+import hashlib
 import json
 import os
 import pickle
@@ -9,7 +18,10 @@ import sys
 import tarfile
 import zipfile
 
+import PIL
+import pkg_resources
 import six
+from keras.utils import Progbar
 from six.moves.urllib.error import HTTPError, URLError
 
 stderr = sys.stderr
@@ -24,11 +36,8 @@ from PIL import Image, ExifTags
 from keras.layers import Dense, Activation
 from keras.models import Model
 from keras.preprocessing import image
-from skimage.transform import resize
-from sklearn.metrics import pairwise_distances
 
-
-from .align_face import align, aligner
+from fawkes.align_face import align
 from six.moves.urllib.request import urlopen
 
 if sys.version_info[0] == 2:
@@ -64,9 +73,24 @@ def clip_img(X, preprocessing='raw'):
     return X
 
 
+IMG_SIZE = 112
+PREPROCESS = 'raw'
+
+
 def load_image(path):
-    img = Image.open(path)
-    if img._getexif() is not None:
+    try:
+        img = Image.open(path)
+    except PIL.UnidentifiedImageError:
+        return None
+    except IsADirectoryError:
+        return None
+
+    try:
+        info = img._getexif()
+    except OSError:
+        return None
+
+    if info is not None:
         for orientation in ExifTags.TAGS.keys():
             if ExifTags.TAGS[orientation] == 'Orientation':
                 break
@@ -87,49 +111,93 @@ def load_image(path):
     return image_array
 
 
+def filter_image_paths(image_paths):
+    print("Identify {} files in the directory".format(len(image_paths)))
+    new_image_paths = []
+    new_images = []
+    for p in image_paths:
+        img = load_image(p)
+        if img is None:
+            print("{} is not an image file, skipped".format(p.split("/")[-1]))
+            continue
+        new_image_paths.append(p)
+        new_images.append(img)
+    print("Identify {} images in the directory".format(len(new_image_paths)))
+    return new_image_paths, new_images
+
+
 class Faces(object):
-    def __init__(self, image_paths, sess, verbose=1):
+    def __init__(self, image_paths, loaded_images, aligner, verbose=1, eval_local=False, preprocessing=True,
+                 no_align=False):
+        self.image_paths = image_paths
         self.verbose = verbose
-        self.aligner = aligner(sess)
+        self.no_align = no_align
+        self.aligner = aligner
+        self.margin = 30
         self.org_faces = []
         self.cropped_faces = []
         self.cropped_faces_shape = []
         self.cropped_index = []
+        self.start_end_ls = []
         self.callback_idx = []
-        if verbose:
-            print("Identify {} images".format(len(image_paths)))
-        for i, p in enumerate(image_paths):
-            cur_img = load_image(p)
+        self.images_without_face = []
+        for i in range(0, len(loaded_images)):
+            cur_img = loaded_images[i]
+            p = image_paths[i]
             self.org_faces.append(cur_img)
-            align_img = align(cur_img, self.aligner, margin=0.7)
-            cur_faces = align_img[0]
 
+            if not no_align:
+                align_img = align(cur_img, self.aligner)
+                if align_img is None:
+                    print("Find 0 face(s) in {}".format(p.split("/")[-1]))
+                    self.images_without_face.append(i)
+                    continue
+
+                cur_faces = align_img[0]
+            else:
+                cur_faces = [cur_img]
+
+            cur_faces = [face for face in cur_faces if face.shape[0] != 0 and face.shape[1] != 0]
             cur_shapes = [f.shape[:-1] for f in cur_faces]
 
             cur_faces_square = []
-            if verbose:
+            if verbose and not no_align:
                 print("Find {} face(s) in {}".format(len(cur_faces), p.split("/")[-1]))
+            if eval_local:
+                cur_faces = cur_faces[:1]
 
             for img in cur_faces:
-                long_size = max([img.shape[1], img.shape[0]])
-                base = np.zeros((long_size, long_size, 3))
-                base[0:img.shape[0], 0:img.shape[1], :] = img
+                if eval_local:
+                    base = resize(img, (IMG_SIZE, IMG_SIZE))
+                else:
+                    long_size = max([img.shape[1], img.shape[0]]) + self.margin
+
+                    base = np.ones((long_size, long_size, 3)) * np.mean(img, axis=(0, 1))
+
+                    start1, end1 = get_ends(long_size, img.shape[0])
+                    start2, end2 = get_ends(long_size, img.shape[1])
+
+                    base[start1:end1, start2:end2, :] = img
+                    cur_start_end = (start1, end1, start2, end2)
+                    self.start_end_ls.append(cur_start_end)
+
                 cur_faces_square.append(base)
-
-            cur_index = align_img[1]
-            cur_faces_square = [resize(f, (224, 224)) for f in cur_faces_square]
-            self.cropped_faces_shape.extend(cur_shapes)
+            cur_faces_square = [resize(f, (IMG_SIZE, IMG_SIZE)) for f in cur_faces_square]
             self.cropped_faces.extend(cur_faces_square)
-            self.cropped_index.extend(cur_index)
-            self.callback_idx.extend([i] * len(cur_faces_square))
 
-        if not self.cropped_faces:
-            print("No faces detected")
-            exit(1)
+            if not self.no_align:
+                cur_index = align_img[1]
+                self.cropped_faces_shape.extend(cur_shapes)
+                self.cropped_index.extend(cur_index[:len(cur_faces_square)])
+                self.callback_idx.extend([i] * len(cur_faces_square))
+
+        if len(self.cropped_faces) == 0:
+            return
 
         self.cropped_faces = np.array(self.cropped_faces)
 
-        self.cropped_faces = preprocess(self.cropped_faces, 'imagenet')
+        if preprocessing:
+            self.cropped_faces = preprocess(self.cropped_faces, PREPROCESS)
 
         self.cloaked_cropped_faces = None
         self.cloaked_faces = np.copy(self.org_faces)
@@ -137,22 +205,41 @@ class Faces(object):
     def get_faces(self):
         return self.cropped_faces
 
-    def merge_faces(self, cloaks):
+    def merge_faces(self, protected_images, original_images):
+        if self.no_align:
+            return np.clip(protected_images, 0.0, 255.0), self.images_without_face
 
         self.cloaked_faces = np.copy(self.org_faces)
 
         for i in range(len(self.cropped_faces)):
-            cur_cloak = cloaks[i]
+            cur_protected = protected_images[i]
+            cur_original = original_images[i]
+
             org_shape = self.cropped_faces_shape[i]
-            old_square_shape = max([org_shape[0], org_shape[1]])
-            reshape_cloak = resize(cur_cloak, (old_square_shape, old_square_shape))
-            reshape_cloak = reshape_cloak[0:org_shape[0], 0:org_shape[1], :]
+
+            old_square_shape = max([org_shape[0], org_shape[1]]) + self.margin
+
+            cur_protected = resize(cur_protected, (old_square_shape, old_square_shape))
+            cur_original = resize(cur_original, (old_square_shape, old_square_shape))
+
+            start1, end1, start2, end2 = self.start_end_ls[i]
+
+            reshape_cloak = cur_protected - cur_original
+            reshape_cloak = reshape_cloak[start1:end1, start2:end2, :]
 
             callback_id = self.callback_idx[i]
             bb = self.cropped_index[i]
-            self.cloaked_faces[callback_id][bb[1]:bb[3], bb[0]:bb[2], :] += reshape_cloak
+            self.cloaked_faces[callback_id][bb[0]:bb[2], bb[1]:bb[3], :] += reshape_cloak
 
-        return self.cloaked_faces
+        for i in range(0, len(self.cloaked_faces)):
+            self.cloaked_faces[i] = np.clip(self.cloaked_faces[i], 0.0, 255.0)
+        return self.cloaked_faces, self.images_without_face
+
+
+def get_ends(longsize, window):
+    start = (longsize - window) // 2
+    end = start + window
+    return start, end
 
 
 def dump_dictionary_as_json(dict, outfile):
@@ -161,6 +248,48 @@ def dump_dictionary_as_json(dict, outfile):
         f.write(j.encode())
 
 
+def load_victim_model(number_classes, teacher_model=None, end2end=False):
+    for l in teacher_model.layers:
+        l.trainable = end2end
+    x = teacher_model.layers[-1].output
+
+    x = Dense(number_classes)(x)
+    x = Activation('softmax', name="act")(x)
+    model = Model(teacher_model.input, x)
+    opt = keras.optimizers.Adadelta()
+    model.compile(loss='categorical_crossentropy', optimizer=opt, metrics=['accuracy'])
+    return model
+
+
+def resize(img, sz):
+    assert np.min(img) >= 0 and np.max(img) <= 255.0
+    from keras.preprocessing import image
+    im_data = image.array_to_img(img).resize((sz[1], sz[0]))
+    im_data = image.img_to_array(im_data)
+    return im_data
+
+
+def init_gpu(gpu):
+    ''' code to initialize gpu in tf2'''
+    if isinstance(gpu, list):
+        gpu_num = ','.join([str(i) for i in gpu])
+    else:
+        gpu_num = str(gpu)
+    if "CUDA_VISIBLE_DEVICES" in os.environ:
+        print('GPU already initiated')
+        return
+    os.environ["CUDA_VISIBLE_DEVICES"] = gpu_num
+    gpus = tf.config.experimental.list_physical_devices('GPU')
+    if gpus:
+        try:
+            tf.config.experimental.set_visible_devices(gpus[0], 'GPU')
+            tf.config.experimental.set_memory_growth(gpus[0], True)
+            logical_gpus = tf.config.experimental.list_logical_devices('GPU')
+            print(len(gpus), "Physical GPUs,", len(logical_gpus), "Logical GPU")
+        except RuntimeError as e:
+            print(e)
+
+
 def fix_gpu_memory(mem_fraction=1):
     os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
     tf_config = None
@@ -176,37 +305,12 @@ def fix_gpu_memory(mem_fraction=1):
     return sess
 
 
-def load_victim_model(number_classes, teacher_model=None, end2end=False):
-    for l in teacher_model.layers:
-        l.trainable = end2end
-    x = teacher_model.layers[-1].output
-    x = Dense(number_classes)(x)
-    x = Activation('softmax', name="act")(x)
-    model = Model(teacher_model.input, x)
-    opt = keras.optimizers.Adadelta()
-    model.compile(loss='categorical_crossentropy', optimizer=opt, metrics=['accuracy'])
-    return model
-
-
-def init_gpu(gpu_index, force=False):
-    if isinstance(gpu_index, list):
-        gpu_num = ','.join([str(i) for i in gpu_index])
-    else:
-        gpu_num = str(gpu_index)
-    if "CUDA_VISIBLE_DEVICES" in os.environ and os.environ["CUDA_VISIBLE_DEVICES"] and not force:
-        print('GPU already initiated')
-        return
-    os.environ["CUDA_VISIBLE_DEVICES"] = gpu_num
-    sess = fix_gpu_memory()
-    return sess
-
-
 def preprocess(X, method):
     assert method in {'raw', 'imagenet', 'inception', 'mnist'}
 
-    if method is 'raw':
+    if method == 'raw':
         pass
-    elif method is 'imagenet':
+    elif method == 'imagenet':
         X = imagenet_preprocessing(X)
     else:
         raise Exception('unknown method %s' % method)
@@ -217,9 +321,9 @@ def preprocess(X, method):
 def reverse_preprocess(X, method):
     assert method in {'raw', 'imagenet', 'inception', 'mnist'}
 
-    if method is 'raw':
+    if method == 'raw':
         pass
-    elif method is 'imagenet':
+    elif method == 'imagenet':
         X = imagenet_reverse_preprocessing(X)
     else:
         raise Exception('unknown method %s' % method)
@@ -320,34 +424,34 @@ def build_bottleneck_model(model, cut_off):
 
 
 def load_extractor(name):
-    model_dir = os.path.join(os.path.expanduser('~'), '.fawkes')
+    hash_map = {"extractor_2": "ce703d481db2b83513bbdafa27434703",
+                "extractor_0": "94854151fd9077997d69ceda107f9c6b"}
+    assert name in ["extractor_2", 'extractor_0']
+    model_file = pkg_resources.resource_filename("fawkes", "model/{}.h5".format(name))
+    cur_hash = hash_map[name]
+    model_dir = pkg_resources.resource_filename("fawkes", "model/")
     os.makedirs(model_dir, exist_ok=True)
-    model_file = os.path.join(model_dir, "{}.h5".format(name))
-    emb_file = os.path.join(model_dir, "{}_emb.p.gz".format(name))
-    if os.path.exists(model_file):
-        model = keras.models.load_model(model_file)
-    else:
-        get_file("{}.h5".format(name), "http://sandlab.cs.uchicago.edu/fawkes/files/{}.h5".format(name),
-                 cache_dir=model_dir, cache_subdir='')
-        model = keras.models.load_model(model_file)
+    get_file("{}.h5".format(name), "http://mirror.cs.uchicago.edu/fawkes/files/{}.h5".format(name),
+             cache_dir=model_dir, cache_subdir='', md5_hash=cur_hash)
 
-    if not os.path.exists(emb_file):
-        get_file("{}_emb.p.gz".format(name), "http://sandlab.cs.uchicago.edu/fawkes/files/{}_emb.p.gz".format(name),
-                 cache_dir=model_dir, cache_subdir='')
-
-    if hasattr(model.layers[-1], "activation") and model.layers[-1].activation == "softmax":
-        raise Exception(
-            "Given extractor's last layer is softmax, need to remove the top layers to make it into a feature extractor")
-    # if "extract" in name.split("/")[-1]:
-    #     pass
-    # else:
-    #     print("Convert a model to a feature extractor")
-    #     model = build_bottleneck_model(model, model.layers[layer_idx].name)
-    #     model.save(name + "extract")
-    #     model = keras.models.load_model(name + "extract")
+    model = keras.models.load_model(model_file)
+    model = Extractor(model)
     return model
 
 
+class Extractor(object):
+    def __init__(self, model):
+        self.model = model
+
+    def predict(self, imgs):
+        imgs = imgs / 255.0
+        embeds = l2_norm(self.model(imgs))
+        return embeds
+
+    def __call__(self, x):
+        return self.predict(x)
+
+
 def get_dataset_path(dataset):
     model_dir = os.path.join(os.path.expanduser('~'), '.fawkes')
     if not os.path.exists(os.path.join(model_dir, "config.json")):
@@ -362,44 +466,20 @@ def get_dataset_path(dataset):
         'num_images']
 
 
-def normalize(x):
-    return x / np.linalg.norm(x, axis=1, keepdims=True)
-
-
 def dump_image(x, filename, format="png", scale=False):
-    # img = image.array_to_img(x, scale=scale)
-    img = image.array_to_img(x)
+    img = image.array_to_img(x, scale=scale)
     img.save(filename, format)
     return
 
 
-def load_dir(path):
-    assert os.path.exists(path)
-    x_ls = []
-    for file in os.listdir(path):
-        cur_path = os.path.join(path, file)
-        im = image.load_img(cur_path, target_size=(224, 224))
-        im = image.img_to_array(im)
-        x_ls.append(im)
-    raw_x = np.array(x_ls)
-    return preprocess(raw_x, 'imagenet')
-
-
 def load_embeddings(feature_extractors_names):
     model_dir = os.path.join(os.path.expanduser('~'), '.fawkes')
-    dictionaries = []
     for extractor_name in feature_extractors_names:
         fp = gzip.open(os.path.join(model_dir, "{}_emb.p.gz".format(extractor_name)), 'rb')
         path2emb = pickle.load(fp)
         fp.close()
 
-        dictionaries.append(path2emb)
-
-    merge_dict = {}
-    for k in dictionaries[0].keys():
-        cur_emb = [dic[k] for dic in dictionaries]
-        merge_dict[k] = np.concatenate(cur_emb)
-    return merge_dict
+    return path2emb
 
 
 def extractor_ls_predict(feature_extractors_ls, X):
@@ -408,22 +488,23 @@ def extractor_ls_predict(feature_extractors_ls, X):
         cur_features = extractor.predict(X)
         feature_ls.append(cur_features)
     concated_feature_ls = np.concatenate(feature_ls, axis=1)
-    concated_feature_ls = normalize(concated_feature_ls)
     return concated_feature_ls
 
 
-def calculate_dist_score(a, b, feature_extractors_ls, metric='l2'):
-    features1 = extractor_ls_predict(feature_extractors_ls, a)
-    features2 = extractor_ls_predict(feature_extractors_ls, b)
+def pairwise_l2_distance(A, B):
+    BT = B.transpose()
+    vecProd = np.dot(A, BT)
+    SqA = A ** 2
+    sumSqA = np.matrix(np.sum(SqA, axis=1))
+    sumSqAEx = np.tile(sumSqA.transpose(), (1, vecProd.shape[1]))
 
-    pair_cos = pairwise_distances(features1, features2, metric)
-    max_sum = np.min(pair_cos, axis=0)
-    max_sum_arg = np.argsort(max_sum)[::-1]
-    max_sum_arg = max_sum_arg[:len(a)]
-    max_sum = [max_sum[i] for i in max_sum_arg]
-    paired_target_X = [b[j] for j in max_sum_arg]
-    paired_target_X = np.array(paired_target_X)
-    return np.min(max_sum), paired_target_X
+    SqB = B ** 2
+    sumSqB = np.sum(SqB, axis=1)
+    sumSqBEx = np.tile(sumSqB, (vecProd.shape[0], 1))
+    SqED = sumSqBEx + sumSqAEx - 2 * vecProd
+    SqED[SqED < 0] = 0.0
+    ED = np.sqrt(SqED)
+    return ED
 
 
 def select_target_label(imgs, feature_extractors_ls, feature_extractors_names, metric='l2'):
@@ -432,34 +513,44 @@ def select_target_label(imgs, feature_extractors_ls, feature_extractors_names, m
     original_feature_x = extractor_ls_predict(feature_extractors_ls, imgs)
 
     path2emb = load_embeddings(feature_extractors_names)
-    items = list(path2emb.items())
+
+    items = list([(k, v) for k, v in path2emb.items()])
     paths = [p[0] for p in items]
     embs = [p[1] for p in items]
     embs = np.array(embs)
 
-    pair_dist = pairwise_distances(original_feature_x, embs, metric)
+    pair_dist = pairwise_l2_distance(original_feature_x, embs)
+    pair_dist = np.array(pair_dist)
+
     max_sum = np.min(pair_dist, axis=0)
-    max_id = np.argmax(max_sum)
+    max_id_ls = np.argsort(max_sum)[::-1]
+
+    max_id = random.choice(max_id_ls[:20])
 
     target_data_id = paths[int(max_id)]
+    print("target ID: {}".format(target_data_id))
+
     image_dir = os.path.join(model_dir, "target_data/{}".format(target_data_id))
-    # if not os.path.exists(image_dir):
+
     os.makedirs(os.path.join(model_dir, "target_data"), exist_ok=True)
     os.makedirs(image_dir, exist_ok=True)
     for i in range(10):
         if os.path.exists(os.path.join(model_dir, "target_data/{}/{}.jpg".format(target_data_id, i))):
             continue
-        get_file("{}.jpg".format(i),
-                 "http://sandlab.cs.uchicago.edu/fawkes/files/target_data/{}/{}.jpg".format(target_data_id, i),
-                 cache_dir=model_dir, cache_subdir='target_data/{}/'.format(target_data_id))
+        try:
+            get_file("{}.jpg".format(i),
+                     "http://mirror.cs.uchicago.edu/fawkes/files/target_data/{}/{}.jpg".format(target_data_id, i),
+                     cache_dir=model_dir, cache_subdir='target_data/{}/'.format(target_data_id))
+        except Exception:
+            pass
 
     image_paths = glob.glob(image_dir + "/*.jpg")
 
     target_images = [image.img_to_array(image.load_img(cur_path)) for cur_path in
                      image_paths]
 
-    target_images = np.array([resize(x, (224, 224)) for x in target_images])
-    target_images = preprocess(target_images, 'imagenet')
+    target_images = np.array([resize(x, (IMG_SIZE, IMG_SIZE)) for x in target_images])
+    target_images = preprocess(target_images, PREPROCESS)
 
     target_images = list(target_images)
     while len(target_images) < len(imgs):
@@ -469,6 +560,18 @@ def select_target_label(imgs, feature_extractors_ls, feature_extractors_names, m
     return np.array(target_images)
 
 
+def l2_norm(x, axis=1):
+    """l2 norm"""
+    norm = tf.norm(x, axis=axis, keepdims=True)
+    output = x / norm
+    return output
+
+
+""" TensorFlow implementation get_file
+https://github.com/tensorflow/tensorflow/blob/v2.3.0/tensorflow/python/keras/utils/data_utils.py#L168-L297
+"""
+
+
 def get_file(fname,
              origin,
              untar=False,
@@ -490,6 +593,8 @@ def get_file(fname,
     datadir = os.path.join(datadir_base, cache_subdir)
     _makedirs_exist_ok(datadir)
 
+    # fname = path_to_string(fname)
+
     if untar:
         untar_fpath = os.path.join(datadir, fname)
         fpath = untar_fpath + '.tar.gz'
@@ -497,12 +602,35 @@ def get_file(fname,
         fpath = os.path.join(datadir, fname)
 
     download = False
-    if not os.path.exists(fpath):
+    if os.path.exists(fpath):
+        # File found; verify integrity if a hash was provided.
+        if file_hash is not None:
+            if not validate_file(fpath, file_hash, algorithm=hash_algorithm):
+                print('A local file was found, but it seems to be '
+                      'incomplete or outdated because the ' + hash_algorithm +
+                      ' file hash does not match the original value of ' + file_hash +
+                      ' so we will re-download the data.')
+                download = True
+    else:
         download = True
 
     if download:
+        print('Downloading data from', origin)
+
+        class ProgressTracker(object):
+            # Maintain progbar for the lifetime of download.
+            # This design was chosen for Python 2.7 compatibility.
+            progbar = None
+
+        def dl_progress(count, block_size, total_size):
+            if ProgressTracker.progbar is None:
+                if total_size == -1:
+                    total_size = None
+                ProgressTracker.progbar = Progbar(total_size)
+            else:
+                ProgressTracker.progbar.update(count * block_size)
+
         error_msg = 'URL fetch failure on {}: {} -- {}'
-        dl_progress = None
         try:
             try:
                 urlretrieve(origin, fpath, dl_progress)
@@ -514,7 +642,7 @@ def get_file(fname,
             if os.path.exists(fpath):
                 os.remove(fpath)
             raise
-        # ProgressTracker.progbar = None
+        ProgressTracker.progbar = None
 
     if untar:
         if not os.path.exists(untar_fpath):
@@ -569,151 +697,52 @@ def _makedirs_exist_ok(datadir):
     else:
         os.makedirs(datadir, exist_ok=True)  # pylint: disable=unexpected-keyword-arg
 
-# class CloakData(object):
-#     def __init__(self, protect_directory=None, img_shape=(224, 224)):
-#
-#         self.img_shape = img_shape
-#         # self.train_data_dir, self.test_data_dir, self.number_classes, self.number_samples = get_dataset_path(dataset)
-#         # self.all_labels = sorted(list(os.listdir(self.train_data_dir)))
-#         self.protect_directory = protect_directory
-#
-#         self.protect_X = self.load_label_data(self.protect_directory)
-#
-#         self.cloaked_protect_train_X = None
-#
-#         self.label2path_train, self.label2path_test, self.path2idx = self.build_data_mapping()
-#         self.all_training_path = self.get_all_data_path(self.label2path_train)
-#         self.all_test_path = self.get_all_data_path(self.label2path_test)
-#         self.protect_class_path = self.get_class_image_files(os.path.join(self.train_data_dir, self.protect_class))
-#
-#     def get_class_image_files(self, path):
-#         return [os.path.join(path, f) for f in os.listdir(path)]
-#
-#     def extractor_ls_predict(self, feature_extractors_ls, X):
-#         feature_ls = []
-#         for extractor in feature_extractors_ls:
-#             cur_features = extractor.predict(X)
-#             feature_ls.append(cur_features)
-#         concated_feature_ls = np.concatenate(feature_ls, axis=1)
-#         concated_feature_ls = normalize(concated_feature_ls)
-#         return concated_feature_ls
-#
-#     def load_embeddings(self, feature_extractors_names):
-#         dictionaries = []
-#         for extractor_name in feature_extractors_names:
-#             path2emb = pickle.load(open("../feature_extractors/embeddings/{}_emb_norm.p".format(extractor_name), "rb"))
-#             dictionaries.append(path2emb)
-#
-#         merge_dict = {}
-#         for k in dictionaries[0].keys():
-#             cur_emb = [dic[k] for dic in dictionaries]
-#             merge_dict[k] = np.concatenate(cur_emb)
-#         return merge_dict
-#
-#     def select_target_label(self, feature_extractors_ls, feature_extractors_names, metric='l2'):
-#         original_feature_x = self.extractor_ls_predict(feature_extractors_ls, self.protect_train_X)
-#
-#         path2emb = self.load_embeddings(feature_extractors_names)
-#         items = list(path2emb.items())
-#         paths = [p[0] for p in items]
-#         embs = [p[1] for p in items]
-#         embs = np.array(embs)
-#
-#         pair_dist = pairwise_distances(original_feature_x, embs, metric)
-#         max_sum = np.min(pair_dist, axis=0)
-#         sorted_idx = np.argsort(max_sum)[::-1]
-#
-#         highest_num = 0
-#         paired_target_X = None
-#         final_target_class_path = None
-#         for idx in sorted_idx[:5]:
-#             target_class_path = paths[idx]
-#             cur_target_X = self.load_dir(target_class_path)
-#             cur_target_X = np.concatenate([cur_target_X, cur_target_X, cur_target_X])
-#             cur_tot_sum, cur_paired_target_X = self.calculate_dist_score(self.protect_train_X, cur_target_X,
-#                                                                          feature_extractors_ls,
-#                                                                          metric=metric)
-#             if cur_tot_sum > highest_num:
-#                 highest_num = cur_tot_sum
-#                 paired_target_X = cur_paired_target_X
-#                 final_target_class_path = target_class_path
-#
-#         np.random.shuffle(paired_target_X)
-#         return final_target_class_path, paired_target_X
-#
-#     def calculate_dist_score(self, a, b, feature_extractors_ls, metric='l2'):
-#         features1 = self.extractor_ls_predict(feature_extractors_ls, a)
-#         features2 = self.extractor_ls_predict(feature_extractors_ls, b)
-#
-#         pair_cos = pairwise_distances(features1, features2, metric)
-#         max_sum = np.min(pair_cos, axis=0)
-#         max_sum_arg = np.argsort(max_sum)[::-1]
-#         max_sum_arg = max_sum_arg[:len(a)]
-#         max_sum = [max_sum[i] for i in max_sum_arg]
-#         paired_target_X = [b[j] for j in max_sum_arg]
-#         paired_target_X = np.array(paired_target_X)
-#         return np.min(max_sum), paired_target_X
-#
-#     def get_all_data_path(self, label2path):
-#         all_paths = []
-#         for k, v in label2path.items():
-#             cur_all_paths = [os.path.join(k, cur_p) for cur_p in v]
-#             all_paths.extend(cur_all_paths)
-#         return all_paths
-#
-#     def load_label_data(self, label):
-#         train_label_path = os.path.join(self.train_data_dir, label)
-#         test_label_path = os.path.join(self.test_data_dir, label)
-#         train_X = self.load_dir(train_label_path)
-#         test_X = self.load_dir(test_label_path)
-#         return train_X, test_X
-#
-#     def load_dir(self, path):
-#         assert os.path.exists(path)
-#         x_ls = []
-#         for file in os.listdir(path):
-#             cur_path = os.path.join(path, file)
-#             im = image.load_img(cur_path, target_size=self.img_shape)
-#             im = image.img_to_array(im)
-#             x_ls.append(im)
-#         raw_x = np.array(x_ls)
-#         return preprocess_input(raw_x)
-#
-#     def build_data_mapping(self):
-#         label2path_train = {}
-#         label2path_test = {}
-#         idx = 0
-#         path2idx = {}
-#         for label_name in self.all_labels:
-#             full_path_train = os.path.join(self.train_data_dir, label_name)
-#             full_path_test = os.path.join(self.test_data_dir, label_name)
-#             label2path_train[full_path_train] = list(os.listdir(full_path_train))
-#             label2path_test[full_path_test] = list(os.listdir(full_path_test))
-#             for img_file in os.listdir(full_path_train):
-#                 path2idx[os.path.join(full_path_train, img_file)] = idx
-#             for img_file in os.listdir(full_path_test):
-#                 path2idx[os.path.join(full_path_test, img_file)] = idx
-#             idx += 1
-#         return label2path_train, label2path_test, path2idx
-#
-#     def generate_data_post_cloak(self, sybil=False):
-#         assert self.cloaked_protect_train_X is not None
-#         while True:
-#             batch_X = []
-#             batch_Y = []
-#             cur_batch_path = random.sample(self.all_training_path, 32)
-#             for p in cur_batch_path:
-#                 cur_y = self.path2idx[p]
-#                 if p in self.protect_class_path:
-#                     cur_x = random.choice(self.cloaked_protect_train_X)
-#                 elif sybil and (p in self.sybil_class):
-#                     cur_x = random.choice(self.cloaked_sybil_train_X)
-#                 else:
-#                     im = image.load_img(p, target_size=self.img_shape)
-#                     im = image.img_to_array(im)
-#                     cur_x = preprocess_input(im)
-#                 batch_X.append(cur_x)
-#                 batch_Y.append(cur_y)
-#             batch_X = np.array(batch_X)
-#             batch_Y = to_categorical(np.array(batch_Y), num_classes=self.number_classes)
-#             yield batch_X, batch_Y
+
+def validate_file(fpath, file_hash, algorithm='auto', chunk_size=65535):
+    """Validates a file against a sha256 or md5 hash.
+    Arguments:
+        fpath: path to the file being validated
+        file_hash:  The expected hash string of the file.
+            The sha256 and md5 hash algorithms are both supported.
+        algorithm: Hash algorithm, one of 'auto', 'sha256', or 'md5'.
+            The default 'auto' detects the hash algorithm in use.
+        chunk_size: Bytes to read at a time, important for large files.
+    Returns:
+        Whether the file is valid
+    """
+    if (algorithm == 'sha256') or (algorithm == 'auto' and len(file_hash) == 64):
+        hasher = 'sha256'
+    else:
+        hasher = 'md5'
+
+    if str(_hash_file(fpath, hasher, chunk_size)) == str(file_hash):
+        return True
+    else:
+        return False
+
+
+def _hash_file(fpath, algorithm='sha256', chunk_size=65535):
+    """Calculates a file sha256 or md5 hash.
+    Example:
+    ```python
+    _hash_file('/path/to/file.zip')
+    'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855'
+    ```
+    Arguments:
+        fpath: path to the file being validated
+        algorithm: hash algorithm, one of `'auto'`, `'sha256'`, or `'md5'`.
+            The default `'auto'` detects the hash algorithm in use.
+        chunk_size: Bytes to read at a time, important for large files.
+    Returns:
+        The file hash
+    """
+    if (algorithm == 'sha256') or (algorithm == 'auto' and len(hash) == 64):
+        hasher = hashlib.sha256()
+    else:
+        hasher = hashlib.md5()
+
+    with open(fpath, 'rb') as fpath_file:
+        for chunk in iter(lambda: fpath_file.read(chunk_size), b''):
+            hasher.update(chunk)
+
+    return hasher.hexdigest()
